# coding=utf8

import audiogen
from itertools import izip_longest

unit = audiogen.util.constant(1)

def setup_mux():
	return audiogen.util.mixer((iter(range(10)),), ((unit,),(unit,)))

def test_mux_output_channels():
	output = setup_mux()
	assert len(output) == 2

def test_mux_duplicated_content():
	output = setup_mux()
	saved = [list(channel) for channel in output]

	# print out first for debugging if test fails
	for num, channel in enumerate(saved):
		print("Channel {}: {}".format(num, ", ".join([str(c) for c in channel])))

	for channel in saved:
		for sample, test in izip_longest(channel, range(10)):
			assert(sample == test)





import logging
logger = logging.getLogger(__name__)

import struct
import wave
import itertools

try:
	import pyaudio
	pyaudio_loaded = True
except ImportError:
	pyaudio_loaded = False
import time
import sys
import errno
import contextlib

from StringIO import StringIO

from .util import hard_clip
from .util import normalize

### Constants

# frame rate in Hertz
FRAME_RATE=44100

# sample width in bytes, 2 = 16 bit
SAMPLE_WIDTH=2
COMPRESSION_TYPE='NONE'
COMPRESSION_NAME='no compression'
BUFFER_SIZE=100000

class frame_rate(object):
	def __init__(self, new_frame_rate):
		global FRAME_RATE
		self._saved_frame_rate = FRAME_RATE
		FRAME_RATE = new_frame_rate
	def __enter__(self, ):
		pass
	def __exit__(self, *args, **kwargs):
		global FRAME_RATE
		FRAME_RATE = self._saved_frame_rate

@contextlib.contextmanager
def sample_width(new_sample_width):
	global SAMPLE_WIDTH
	saved_sample_width = SAMPLE_WIDTH
	SAMPLE_WIDTH = new_sample_width
	yield
	SAMPLE_WIDTH = saved_sample_width

def file_is_seekable(f):
	'''
	Returns True if file `f` is seekable, and False if not
	
	Useful to determine, for example, if `f` is STDOUT to 
	a pipe.
	'''
	try:
		f.tell()
		logger.info("File is seekable!")
	except IOError, e:
		if e.errno == errno.ESPIPE:
			return False
		else:
			raise
	return True


def sample(generator, min=-1, max=1, width=SAMPLE_WIDTH):
	'''Convert audio waveform generator into packed sample generator.'''
	# select signed char, short, or in based on sample width
	fmt = { 1: '<B', 2: '<h', 4: '<i' }[width]
	return (struct.pack(fmt, int(sample)) for sample in \
			normalize(hard_clip(generator, min, max),\
				min, max, -2**(width * 8 - 1), 2**(width * 8 - 1) - 1))
#	scale = float(2**(width * 8) - 1) / (max - min)
#	return (struct.pack('h', int((sample - min) * scale) - 2**(width * 8 - 1)) for sample in generator)

def sample_all(generators, *args, **kwargs):
	'''Convert list of audio waveform generators into list of packed sample generators.'''
	return [sample(gen, *args, **kwargs) for gen in  generators]

def interleave(channels):
	'''
	Interleave samples from multiple channels for wave output

	Accept a list of channel generators and generate a sequence
	of frames, one sample from each channel per frame, in the order
	of the channels in the list. 
	'''
	while True:
		yield "".join([channel.next() for channel in channels])

def buffer(stream, buffer_size=BUFFER_SIZE):
	'''
	Buffer the generator into byte strings of buffer_size samples

	Return a generator that outputs reasonably sized byte strings
	containing buffer_size samples from the generator stream. 

	This allows us to outputing big chunks of the audio stream to 
	disk at once for faster writes.
	'''
	i = iter(stream)
	return iter(lambda: "".join(itertools.islice(i, buffer_size)), "") 
	

def wav_samples(channels, sample_width=SAMPLE_WIDTH, raw_samples=False):
	if hasattr(channels, 'next'):
		# if passed one generator, we have one channel
		channels = (channels,)
	channel_count = len(channels)
	if not raw_samples:
		# we have audio waveforms, so sample/pack them first
		channels = sample_all(channels, width=sample_width)
	return interleave(channels)

class NonSeekableFileProxy(object):
	def __init__(self, file_instance):
		'''Proxy to protect seek and tell methods of non-seekable file objects'''
		self.f = file_instance
	def __getattr__(self, attr):
		def dummy(*args):
			logger.debug("Suppressed call to '{0}'".format(attr))
			return 0
		if attr in ('seek', 'tell'):
			return dummy
		else:
			return getattr(self.f, attr)

def wave_module_patched():
	'''True if wave module can write data size of 0xFFFFFFFF, False otherwise.'''
	f = StringIO()
	w = wave.open(f, "wb")
	w.setparams((1, 2, 44100, 0, "NONE", "no compression"))
	patched = True
	try:
		w.setnframes((0xFFFFFFFF - 36) / w.getnchannels() / w.getsampwidth())
		w._ensure_header_written(0)
	except struct.error:
		patched = False
		logger.info("Error setting wave data size to 0xFFFFFFFF; wave module unpatched, setting sata size to 0x7FFFFFFF")
		w.setnframes((0x7FFFFFFF - 36) / w.getnchannels() / w.getsampwidth())
		w._ensure_header_written(0)
	return patched

def write_wav(f, channels, sample_width=SAMPLE_WIDTH, raw_samples=False, seekable=None):
	stream = wav_samples(channels, sample_width, raw_samples)
	channel_count = 1 if hasattr(channels, "next") else len(channels)

	output_seekable = file_is_seekable(f) if seekable is None else seekable

	if not output_seekable:
		# protect the non-seekable file, since Wave_write will call tell
		f = NonSeekableFileProxy(f)

	w = wave.open(f)
	w.setparams((
		channel_count, 
		sample_width, 
		FRAME_RATE, 
		0, # setting zero frames, should update automatically as more frames written
		COMPRESSION_TYPE, 
		COMPRESSION_NAME
		))

	if not output_seekable:
		if wave_module_patched():
			# set nframes to make wave module write data size of 0xFFFFFFF
			w.setnframes((0xFFFFFFFF - 36) / w.getnchannels() / w.getsampwidth())
			logger.debug("Setting frames to: {0}, {1}".format((w.getnframes()), w._nframes))
		else:
			w.setnframes((0x7FFFFFFF - 36) / w.getnchannels() / w.getsampwidth())
			logger.debug("Setting frames to: {0}, {1}".format((w.getnframes()), w._nframes))
	
	for chunk in buffer(stream):
		logger.debug("Writing %d bytes..." % len(chunk))
		if output_seekable:
			w.writeframes(chunk)
		else:
			# tell wave module not to update nframes header field 
			# if output stream not seekable, e.g. STDOUT to a pipe
			w.writeframesraw(chunk)
	w.close()

def cache_finite_samples(f):
	'''Decorator to cache audio samples produced by the wrapped generator.'''
	cache = {}
	def wrap(*args):
		key = FRAME_RATE, args
		if key not in cache:
			cache[key] = [sample for sample in f(*args)]
		return (sample for sample in cache[key])
	return wrap

def _pyaudio_callback(wavgen):
	def cb(a, frame_count, b, c):
		data = bytearray()
		try:
			for i in range(frame_count):
				data.extend(next(wavgen))
		except StopIteration:
			if len(data) == 0:
				return "", pyaudio.paComplete
		return str(data), pyaudio.paContinue
	return cb

def play(channels, blocking=True, raw_samples=False):
	'''
	Play the contents of the generator using PyAudio

	Play to the system soundcard using PyAudio. PyAudio, an otherwise optional
	depenency, must be installed for this feature to work. 
	'''
	if not pyaudio_loaded:
		raise Exception("Soundcard playback requires PyAudio. Install with `pip install pyaudio`.")

	channel_count = 1 if hasattr(channels, "next") else len(channels)
	wavgen = wav_samples(channels, raw_samples=raw_samples)
	p = pyaudio.PyAudio()
	stream = p.open(
		format=p.get_format_from_width(SAMPLE_WIDTH),
		channels=channel_count,
		rate=FRAME_RATE,
		output=True,
		stream_callback=_pyaudio_callback(wavgen) if not blocking else None
	)
	if blocking:
		try:
			for chunk in buffer(wavgen, 1024):
				stream.write(chunk)
		except Exception:
			raise
		finally:
			if not stream.is_stopped():
				stream.stop_stream()
			try:
				stream.close()
			except Exception:
				pass
	else:
		return stream






from .sampler import frame_rate
from .sampler import write_wav

from .generators import tone
from .generators import beep
from .generators import silence

from .util import crop
from .util import crop_at_zero_crossing
from .util import crop_with_fades
from .util import crop_with_fade_out

from .util import loop
from .util import play





# coding=utf8

'''
Assorted generators 
'''

import math
import itertools

import util
import sampler

TWO_PI = 2 * math.pi

## Audio sample generators

def beep(frequency=440):
	for sample in util.crop_with_fades(tone(frequency), seconds=0.25):
		yield sample

def tone(frequency=440, min_=-1, max_=1):
	def fixed_tone(frequency):
		period = int(sampler.FRAME_RATE / frequency) 
		time_scale = 2 * math.pi / period # period * scale = 2 * pi
		# precompute fixed tone samples # TODO: what about phase glitches at end?
		samples = [math.sin(i * time_scale) for i in xrange(period)]
		while True:
			for i in xrange(period):
				yield samples[i]
	def variable_tone(frequency):
		time_scale = TWO_PI / sampler.FRAME_RATE
		phase = 0
		for f in frequency:
			yield math.sin(phase)

			phase += time_scale * f

			# don't reset hard to zero – avoids sudden phase glitches due to rounding error
			if phase > TWO_PI:
				phase -= TWO_PI

	if not hasattr(frequency, 'next'):
		gen = fixed_tone(frequency)
	else:
		gen = variable_tone(frequency)
	return util.normalize(gen, -1, 1, min_, max_)

def synth(freq, angles):
	if isinstance(angles, (int, float)):
		# argument was just the end angle
		angles = [0, angles]
	gen = tone(freq)
	loop = list(itertools.islice(gen, (sampler.FRAME_RATE / freq) * (angles[1] / (2.0 * math.pi))))[int((sampler.FRAME_RATE / freq) * (angles[0] / (2.0 * math.pi))):]
	while True:
		for sample in loop:
			yield sample

def silence(seconds=None):
	if seconds != None:
		for i in range(int(sampler.FRAME_RATE * seconds)):
			yield 0
	else:
		while True:
			yield 0





import logging 
logger = logging.getLogger(__name__)

import itertools
import struct
import math

from .noise import white_noise
from .noise import white_noise_samples
from .noise import red_noise

import sampler 

def crop(gens, seconds=5, cropper=None):
	'''
	Crop the generator to a finite number of frames

	Return a generator which outputs the provided generator limited
	to enough samples to produce seconds seconds of audio (default 5s)
	at the provided frame rate.
	'''
	if hasattr(gens, "next"):
		# single generator
		gens = (gens,)

	if cropper == None:
		print "cropping %d channel(s) to %d samples each" % (len(gens), seconds * sampler.FRAME_RATE)
		cropper = lambda gen: itertools.islice(gen, 0, seconds * sampler.FRAME_RATE)

	cropped = [cropper(gen) for gen in gens]
	return cropped[0] if len(cropped) == 1 else cropped

def crop_with_fades(gen, seconds, fade_in=0.01, fade_out=0.01):
	source = iter(gen)

	total_samples = int(seconds * sampler.FRAME_RATE)
	fade_in_samples = int(fade_in * sampler.FRAME_RATE)
	fade_out_samples = int(fade_out * sampler.FRAME_RATE)

	start = itertools.islice(source, 0, fade_in_samples)
	middle = itertools.islice(source, 0, total_samples - (fade_in_samples + fade_out_samples))
	end = itertools.islice(source, 0, fade_out_samples)

	def linear_fade(samples, direction="in"):
		for i in xrange(samples):
			if direction == "in":
				yield ( 1.0 / samples ) * i + 0
			elif direction == "out":
				yield ( -1.0 / samples ) * i + 1
			else:
				raise Exception('Fade direction must be "in" or "out"')
		while True:
			yield 0

	for sample in multiply(start, linear_fade(fade_in_samples, direction="in")):
		yield sample

	for sample in middle:
		yield sample
	
	for sample in multiply(end, linear_fade(fade_out_samples, direction="out")):
		yield sample

def crop_with_fade_out(gen, seconds, fade=.01):
	source = iter(gen)

	total_samples = int(seconds * sampler.FRAME_RATE)
	fade_samples = int(fade * sampler.FRAME_RATE)

	start = itertools.islice(source, 0, total_samples - fade_samples)
	end = itertools.islice(source, 0, fade_samples)

	def fader():
		for i in xrange(fade_samples):
			yield 1 - (float(i) / fade_samples) ** 1
		while True:
			yield 0

	for sample in start:
		yield sample
	
	for sample in multiply(end, fader()):
		yield sample


def crop_at_zero_crossing(gen, seconds=5, error=0.1):
	'''
	Crop the generator, ending at a zero-crossing

	Crop the generator to produce approximately seconds seconds 
	(default 5s) of audio at the provided FRAME_RATE, attempting 
	to end the clip at a zero crossing point to avoid clicking. 
	'''
	source = iter(gen)
	buffer_length = int(2 * error * sampler.FRAME_RATE)
	
	# split the source into two iterators:
	# - start, which contains the bulk of the sound clip
	# - and end, which contains the final 100ms, plus 100ms past 
	#   the desired clip length. We may cut the clip anywhere 
	#   within this +/-100ms end buffer.
	start = itertools.islice(source, 0, int((seconds - error) * sampler.FRAME_RATE))
	end = itertools.islice(source, 0, buffer_length)

	for sample in start:
		yield sample

	# pull end buffer generator into memory so we can work with it
	end = list(end)

	# find min by sorting buffer samples, first by abs of sample, then by distance from optimal
	best = sorted(enumerate(end), key=lambda x: (math.fabs(x[1]),abs((buffer_length/2)-x[0])))
	print best[:10]
	print best[0][0]

	# todo: better logic when we don't have a perfect zero crossing
	#if best[0][1] != 0:
	#	# we don't have a perfect zero crossing, so let's look for best fit?
	#	pass
	
	# crop samples at index of best zero crossing
	for sample in end[:best[0][0] + 1]:
		yield sample


def normalize(generator, min_in=0, max_in=256, min_out=-1, max_out=1):
	scale = float(max_out - min_out) / (max_in - min_in)
	return ((sample - min_in) * scale + min_out for sample in generator)

def hard_clip(generator, min=-1, max=1):
	while True:
		sample = generator.next()
		if sample > max:
			logger.warn("Warning, clipped value %f > max %f" % (sample, max))
			yield max
		elif sample < min:
			logger.warn("Warning, clipped value %f < min %f" % (sample, min))
			yield min
		else:
			yield sample

def vector_reduce(op, generators):
	while True:
		yield reduce(op, [g.next() for g in generators])
def vector_reduce1(op, generators):
	while True:
		yield reduce(op, [g.next() for g in generators])

def sum(*generators):
	return vector_reduce(lambda a,b: a + b, generators)

def multiply(*generators):
	return vector_reduce1(lambda a,b: a * b, generators)

def constant(value):
	while True:
		yield value

# filters

def volume(gen, dB=0):
	'''Change the volume of gen by dB decibles'''
	if not hasattr(dB, 'next'):
		# not a generator
		scale = 10 ** (dB / 20.)
	else:
		def scale_gen():
			while True:
				yield 10 ** (next(dB) / 20.)
		scale = scale_gen()
	return envelope(gen, scale)

def clip(gen, limit):
	if not hasattr(limit, 'next'):
		limit = constant(limit)
	while True:
		sample = gen.next()
		current_limit = limit.next()
		if math.fabs(sample) > current_limit:
			yield current_limit * (math.fabs(sample) / sample if sample != 0 else 0)
		else:
			yield sample

def envelope(gen, volume):
	if not hasattr(volume, 'next'):
		volume = constant(volume)
	while True:
		sample = gen.next()
		current_volume = volume.next()
		yield current_volume * sample

def loop(*gens):
	loops = [list(gen) for gen in gens]
	while True:
		for loop in loops: 
			for sample in loop:
				yield sample

def mixer(inputs, mix=None):
	'''
	Mix `inputs` together based on `mix` tuple

	`inputs` should be a tuple of *n* generators. 

	`mix` should be a tuple of *m* tuples, one per desired
	output channel. Each of the *m* tuples should contain
	*n* generators, corresponding to the time-sequence of 
	the desired mix levels for each of the *n* input channels.

	That is, to make an ouput channel contain a 50/50 mix of the
	two input channels, the tuple would be:

	    (constant(0.5), constant(0.5))
	
	The mix generators need not be constant, allowing for time-varying
	mix levels: 

	    # 50% from input 1, pulse input 2 over a two second cycle
	    (constant(0.5), tone(0.5))

	The mixer will return a list of *m* generators, each containing 
	the data from the inputs mixed as specified. 

	If no `mix` tuple is specified, all of the *n* input channels
	will be mixed together into one generator, with the volume of 
	each reduced *n*-fold.

	Example:

	    # three in, two out; 
	    # 10Hz binaural beat with white noise across both channels
	    mixer(
	    		(white_noise(), tone(440), tone(450)), 
	    		(
	    			(constant(.5), constant(1), constant(0)),
	    			(constant(.5), constant(0), constant(1)),
	    		)
	    	)
	'''
	if mix == None:
		# by default, mix all inputs down to one channel
		mix = ([constant(1.0 / len(inputs))] * len(inputs),)

	duped_inputs = zip(*[itertools.tee(i, len(mix)) for i in inputs])

# second zip is backwards
	return [\
			sum(*[multiply(m,i) for m,i in zip(channel_mix, channel_inputs)])\
			for channel_mix, channel_inputs in zip(mix, duped_inputs) \
			]

def channelize(gen, channels):
	'''
	Break multi-channel generator into one sub-generator per channel

	Takes a generator producing n-tuples of samples and returns n generators,
	each producing samples for a single channel.

	Since multi-channel generators are the only reasonable way to synchronize samples
	across channels, and the sampler functions only take tuples of generators,
	you must use this function to process synchronized streams for output.
	'''
	def pick(g, channel):
		for samples in g:
			yield samples[channel]
	return [pick(gen_copy, channel) for channel, gen_copy in enumerate(itertools.tee(gen, channels))]

def play(filename):
	import subprocess
	subprocess.call(["afplay", filename])





import util as util

# Arcfour PRNG as a fast source of repeatable randomish numbers; totally unnecessary here, but simple.
def arcfour(key, csbN=1):
	'''Return a generator for the ARCFOUR/RC4 pseudorandom keystream for the 
	   key provided. Keys should be byte strings or sequences of ints.'''
	if isinstance(key, str):
		key = [ord(c) for c in key]
	s = range(256)
	j = 0
	for n in range(csbN):
		for i in range(256):
			j = (j + s[i] + key[i % len(key)]) % 256
			t = s[i]
			s[i] = s[j]
			s[j] = t
	i = 0
	j = 0
	while True:
		i = (i + 1) % 256
		j = (j + s[i]) % 256
		t = s[i] 
		s[i] = s[j]
		s[j] = t
		yield s[(s[i] + s[j]) % 256]

def arcfour_drop(key, n=3072):
	'''Return a generator for the RC4-drop pseudorandom keystream given by 
	   the key and number of bytes to drop passed as arguments. Dropped bytes
	   default to the more conservative 3072, NOT the SCAN default of 768.'''
	af = arcfour(key)
	[af.next() for c in range(n)]
	return af
mark_4 = arcfour_drop


def white_noise(key=(1,2,3,4,5)):
	def prng(key):
		af = arcfour(key)
		while True:
			yield af.next() * 2**8 + af.next()
	return util.normalize(prng(key), 0, 2**16)

def white_noise_samples(key=(1,2,3,4,5)):
	af = arcfour(key)
	while True:
		yield chr(af.next()) + chr(af.next())

def red_noise(key=(1,2,3,4,5)):
	def random_walk(key, min=-1024, max=1024):
		af = arcfour(key)
		sample = 0
		while True:
			sample += af.next() - 128
			if sample > max:
				sample = max 
			elif sample < min:
				sample = min
			yield sample
	return util.normalize(random_walk(key), -1024, 1024)

'''
RFC 6229
2.  Test Vectors for RC4

 Key length: 40 bits.
 key: 0x0102030405

 DEC    0 HEX    0:  b2 39 63 05  f0 3d c0 27   cc c3 52 4a  0a 11 18 a8
 DEC   16 HEX   10:  69 82 94 4f  18 fc 82 d5   89 c4 03 a4  7a 0d 09 19
 DEC  240 HEX   f0:  28 cb 11 32  c9 6c e2 86   42 1d ca ad  b8 b6 9e ae
 DEC  256 HEX  100:  1c fc f6 2b  03 ed db 64   1d 77 df cf  7f 8d 8c 93
 DEC  496 HEX  1f0:  42 b7 d0 cd  d9 18 a8 a3   3d d5 17 81  c8 1f 40 41
 DEC  512 HEX  200:  64 59 84 44  32 a7 da 92   3c fb 3e b4  98 06 61 f6
 DEC  752 HEX  2f0:  ec 10 32 7b  de 2b ee fd   18 f9 27 76  80 45 7e 22
 DEC  768 HEX  300:  eb 62 63 8d  4f 0b a1 fe   9f ca 20 e0  5b f8 ff 2b
 DEC 1008 HEX  3f0:  45 12 90 48  e6 a0 ed 0b   56 b4 90 33  8f 07 8d a5
 DEC 1024 HEX  400:  30 ab bc c7  c2 0b 01 60   9f 23 ee 2d  5f 6b b7 df
 DEC 1520 HEX  5f0:  32 94 f7 44  d8 f9 79 05   07 e7 0f 62  e5 bb ce ea
 DEC 1536 HEX  600:  d8 72 9d b4  18 82 25 9b   ee 4f 82 53  25 f5 a1 30
 DEC 2032 HEX  7f0:  1e b1 4a 0c  13 b3 bf 47   fa 2a 0b a9  3a d4 5b 8b
 DEC 2048 HEX  800:  cc 58 2f 8b  a9 f2 65 e2   b1 be 91 12  e9 75 d2 d7
 DEC 3056 HEX  bf0:  f2 e3 0f 9b  d1 02 ec bf   75 aa ad e9  bc 35 c4 3c
 DEC 3072 HEX  c00:  ec 0e 11 c4  79 dc 32 9d   c8 da 79 68  fe 96 56 81
 DEC 4080 HEX  ff0:  06 83 26 a2  11 84 16 d2   1f 9d 04 b2  cd 1c a0 50
 DEC 4096 HEX 1000:  ff 25 b5 89  95 99 67 07   e5 1f bd f0  8b 34 d8 75
'''



#!/usr/bin/env python

from setuptools import setup, find_packages

required_modules = []
extras_require = {
	'soundcard_playback': ['pyaudio'],
}

with open("README.rst", "rb") as f:
	readme = f.read().decode('utf8')

setup(
	name="audiogen",
	version="0.1.1",
	description="Generator based tools for working with audio clips.",
	author="Christopher H. Casebeer",
	author_email="",
	url="https://github.com/casebeer/audiogen",

	packages=find_packages(exclude='tests'),
	install_requires=required_modules,
	extras_require=extras_require,

	tests_require=["nose"],
	test_suite="nose.collector",

	long_description=readme,
	classifiers=[
		"License :: OSI Approved :: BSD License",
		"Programming Language :: Python :: 2.6",
		"Programming Language :: Python :: 2.7",
		"Intended Audience :: Developers",
		"Topic :: Multimedia :: Sound/Audio",
	]
)

